/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#if defined(_WIN32) || defined(__CYGWIN__) // Windows

#  undef WINVER
#  define WINVER 0x0A00
#  undef _WIN32_WINNT
#  define _WIN32_WINNT 0x0A00
#  undef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
#  undef DEFINE_CONSOLEV2_PROPERTIES
#  define DEFINE_CONSOLEV2_PROPERTIES
#  include <windows.h>
#  include "win10vt.h"

// some implementations don't define them:
#  ifndef ENABLE_VIRTUAL_TERMINAL_PROCESSING
#    define ENABLE_VIRTUAL_TERMINAL_PROCESSING ((DWORD)0x4)
#  endif
#  ifndef DISABLE_NEWLINE_AUTO_RETURN
#    define DISABLE_NEWLINE_AUTO_RETURN ((DWORD)0x8)
#  endif
#  ifndef ENABLE_VIRTUAL_TERMINAL_INPUT
#    define ENABLE_VIRTUAL_TERMINAL_INPUT ((DWORD)0x200)
#  endif

int TargetIsWindowsConsole(void)
{
  return 0 != SendMessageW(GetConsoleWindow(), WM_GETICON, 0, 0);
}

unsigned SetVTConMode(unsigned f_use, unsigned f_enable)
{
  HANDLE h_in = INVALID_HANDLE_VALUE, h_out = INVALID_HANDLE_VALUE;
  DWORD mode = 0UL;
  unsigned ret = VT_CONNONE;

  if (f_use & VT_CONIN &&
      INVALID_HANDLE_VALUE != (h_in = CreateFileW(L"CONIN$",
                                                  GENERIC_READ | GENERIC_WRITE,
                                                  FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                  NULL,
                                                  OPEN_EXISTING,
                                                  0UL,
                                                  NULL)) &&
      GetConsoleMode(h_in, &mode) &&
      SetConsoleMode(h_in, f_enable & VT_CONIN ? mode | ENABLE_VIRTUAL_TERMINAL_INPUT : mode & ~ENABLE_VIRTUAL_TERMINAL_INPUT))
    ret |= VT_CONIN;

  if (f_use & VT_CONOUT &&
      INVALID_HANDLE_VALUE != (h_out = CreateFileW(L"CONOUT$",
                                                   GENERIC_READ | GENERIC_WRITE,
                                                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                                                   NULL,
                                                   OPEN_EXISTING,
                                                   0UL,
                                                   NULL)) &&
      GetConsoleMode(h_out, &mode) &&
      SetConsoleMode(h_out, f_enable & VT_CONOUT ? mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING | DISABLE_NEWLINE_AUTO_RETURN : mode & ~(ENABLE_VIRTUAL_TERMINAL_PROCESSING | DISABLE_NEWLINE_AUTO_RETURN)))
    ret |= VT_CONOUT;

  if (h_in != INVALID_HANDLE_VALUE)
    CloseHandle(h_in);

  if (h_out != INVALID_HANDLE_VALUE)
    CloseHandle(h_out);

  return ret;
}

unsigned GetStdReferCon()
{
  HANDLE h_in = GetStdHandle(STD_INPUT_HANDLE),
         h_out = GetStdHandle(STD_OUTPUT_HANDLE),
         h_err = GetStdHandle(STD_ERROR_HANDLE);
  unsigned ret = VT_STDNONE;
  INPUT_RECORD inrec = { 0 };
  DWORD cnt = 0UL;

  // Pipes and redirections are almost always "one-ways". Even if Microsoft suggests to use
  // GetConsoleMode to figure out if a console handle is referenced, this would fail here.
  // E.g. you might be writing to a console even if GetConsoleMode wasn't able to read the properties
  // because reading goes into the wrong direction of a pipe or a redirection. Thus, the only reliable
  // way to determine if a console handle was referenced is to actually read/write using the console
  // functions explicitly. It's reasonable to try reading/writing zero information. Functions will
  // still succeed or fail, respectively.
  if (ReadConsoleInputW(h_in, &inrec, 0UL, &cnt))
    ret |= VT_STDIN;

  if (WriteConsoleW(h_out, (void *)&inrec, 0UL, &cnt, NULL))
    ret |= VT_STDOUT;

  if (WriteConsoleW(h_err, (void *)&inrec, 0UL, &cnt, NULL))
    ret |= VT_STDERR;

  return ret;
}

#else // not Windows

#  include "win10vt.h"

int TargetIsWindowsConsole(void)
{
  return 0;
}

unsigned SetVTConMode(unsigned f_use, unsigned f_enable)
{
  f_use = f_use, f_enable = f_enable;
  return 0;
}

unsigned GetStdReferCon()
{
  return 0;
}

#endif
