/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/// @file      win10vt.h
/// @brief     Conditioning of the Windows console to support Virtual Terminal sequences.
///            Identification of standard streams where Virtual Terminal sequences are supported.
/// @version   1.1
/// @author    Steffen Illhardt
/// @date      2022
/// @copyright MIT License
/// @note      Functions are only effective for the Windows console host V2. <br>
///            Use the `TargetIsWindowsConsole()` function to determine whether a Windows console is attached to the process.

#ifndef WIN10VT_H_INCLUDED__
#define WIN10VT_H_INCLUDED__

#ifdef __cplusplus
extern "C" {
#endif

/// @{
/// @brief  Function-like macros to simplify the creation of VT sequences as ANSI or
///          wide string literals, respectively. <br>
///          For supported sequences refer to: <br>
///          https://docs.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences
#define ESC(seq) "\x1B" #seq
#define ESCW(seq) L"\x1B" L## #seq
/// @}

/// @{
/// @brief  Flags representing the console handles. <br>
///         NOTE: VT_CONIN might be referenced by stdin
///               while VT_CONOUT might be referenced by both stdout and stderr.
///               There is no separate console handle for stderr.
#define VT_CONNONE (0U)
#define VT_CONIN (1U)
#define VT_CONOUT (2U)
#define VT_CONALL (VT_CONIN | VT_CONOUT)
/// @}

/// @brief  Determine whether or not the Windows Console Host is targeted.
/// @return  1 if the target is the Windows Console Host, 0 if not.
int TargetIsWindowsConsole(void);

/// @brief  Enable or disable Virtual Terminal processing for the console window.
/// @param  f_use     Flags referencing the handles to be updated. One or more out of VT_CONIN, VT_CONOUT.
/// @param  f_enable  Flags out of f_use where VT processing shall be enabled. VT processing will be
///                   disabled for handles where flags are set in f_use but not in f_enable.
/// @return  Flags which represent the console handles where VT processing was updated.
///          That is, either none (VT_CONNONE), one, or both of VT_CONIN, VT_CONOUT. <br>
///          NOTE: Not being able to update the console handles may indicate that either no console window
///                is attached, or that a legacy console is attached where VT processing is not supported.
unsigned SetVTConMode(unsigned f_use, unsigned f_enable);

/// @{
/// @brief  Flags representing the standard handles. <br>
///         NOTE: VT_STDIN may refer to CONIN$, and
///               both VT_STDOUT and VT_STDERR may refer to CONOUT$.
///               There is no separate console device for VT_STDERR.
#define VT_STDNONE (0U)
#define VT_STDIN (1U)
#define VT_STDOUT (2U)
#define VT_STDERR (4U)
#define VT_STDALL (VT_STDIN | VT_STDOUT | VT_STDERR)
/// @}

/// @brief  Determine which of the standard handles refer to a console.
/// @return  Flags which represent the standard handles that refer to a console.
///          That is, either none (VT_STDNONE), one, or more out of VT_STDIN, VT_STDOUT, VT_STDERR.
unsigned GetStdReferCon(void);

#ifdef __cplusplus
}
#endif
#endif // WIN10VT_H_INCLUDED__
